@@ -129,7 +129,7 @@ w_err(SerdWriter* writer, SerdStatus st, const char* fmt, ...)
 
 	va_list args;
 	va_start(args, fmt);
-	const SerdError e = { st, NULL, 0, 0, fmt, &args };
+	const SerdError e = { st, (const uint8_t*)"", 0, 0, fmt, &args };
 	serd_error(writer->error_sink, writer->error_handle, &e);
 	va_end(args);
 }
@@ -419,7 +419,6 @@ write_literal(SerdWriter*        writer,
               const SerdNode*    node,
               const SerdNode*    datatype,
               const SerdNode*    lang,
-              Field              field,
               SerdStatementFlags flags)
 {
 	if (supports_abbrev(writer) && datatype && datatype->buf) {
@@ -621,7 +620,7 @@ write_node(SerdWriter*        writer,
 	bool ret = false;
 	switch (node->type) {
 	case SERD_LITERAL:
-		ret = write_literal(writer, node, datatype, lang, field, flags);
+		ret = write_literal(writer, node, datatype, lang, flags);
 		break;
 	case SERD_URI:
 		ret = write_uri_node(writer, node, field, flags);
